/*
  Graphic LCD NOKIA3300-LCD Control Library

  T. Nakagawa
  2014/01/09
*/


#include <ftdi.h>
#include <unistd.h>
#include "glcd.h"


#define BAUDRATE 1000000
#define BIT_CS 0x01
#define BIT_SCK 0x02
#define BIT_SDATA 0x04
#define CMD_DISON  0xaf
#define CMD_DISOFF 0xae
#define CMD_DISINV 0xa7
#define CMD_COMSCN 0xbb
#define CMD_DISCTL 0xca
#define CMD_SLPIN 0x95
#define CMD_SLPOUT 0x94
#define CMD_PASET  0x75
#define CMD_CASET  0x15
#define CMD_DATCTL 0xbc
#define CMD_RGBSET8 0xce
#define CMD_RAMWR  0x5c
#define CMD_PTLOUT 0xa9
#define CMD_OSCON  0xd1
#define CMD_OSCOFF 0xd2
#define CMD_PWRCTR 0x20
#define CMD_VOLCTR 0x81


static struct ftdi_context Ftdic;
static unsigned char Buf_mem[132 * 132 * 9 * 2 + 256];
static unsigned char *Buf = Buf_mem + 1;


static void dev_write(int a0, unsigned char val) {
  int i;

  if (a0 != 0) {
    *Buf++ = BIT_SDATA;
    *Buf++ = (BIT_SDATA | BIT_SCK);
  } else {
    *Buf++ = 0;
    *Buf++ = BIT_SCK;
  }
  for (i = 0; i < 8; i++) {
    if ((val & 0x80) != 0) {
      *Buf++ = BIT_SDATA;
      *Buf++ = (BIT_SDATA | BIT_SCK);
    } else {
      *Buf++ = 0;
      *Buf++ = BIT_SCK;
    }
    val <<= 1;
  }
}


static void dev_flush(void) {
  *Buf_mem = 0;  // CS: H->L
  *Buf++ = BIT_CS;  // CS: L->H
  ftdi_write_data(&Ftdic, Buf_mem, Buf - Buf_mem);
  Buf = Buf_mem + 1;
}


int glcd_open(void) {
  int i;

  if (ftdi_init(&Ftdic) < 0) {
    return 1;
  }
  if (ftdi_usb_open(&Ftdic, 0x0403, 0x6001) < 0) {
    return 2;
  }
  if (ftdi_set_baudrate(&Ftdic, BAUDRATE) < 0) {
    return 3;
  }
  if (ftdi_set_bitmode(&Ftdic, 0x0f, BITMODE_BITBANG) < 0) {
    return 4;
  }

  // Display Setting
  dev_write(0, CMD_DISCTL);
  dev_write(1, 0);
  dev_write(1, 32);
  dev_write(1, 11);
  dev_write(1, 0);
  dev_write(0, CMD_COMSCN);
  dev_write(1, 1);
  dev_write(0, CMD_OSCON);
  dev_write(0, CMD_SLPOUT);
  dev_flush();

  // Power Supply Setting
  dev_write(0, CMD_VOLCTR);
  dev_write(1, 33);
  dev_write(1, 3);
  dev_write(0, CMD_PWRCTR);
  dev_write(1, 0x0f);
  dev_flush();

  // Display Setting 2
  dev_write(0, CMD_DISINV);
  dev_write(0, CMD_PTLOUT);
  dev_flush();

  // Display Setting 3
  dev_write(0, CMD_DATCTL);
  dev_write(1, 0x02);
  dev_write(1, 0x01);
  dev_write(1, 0x01);
  dev_write(0, CMD_RGBSET8);
  for (i = 0; i < 8; i++) {
    dev_write(1, 0x0f * i / (8 - 1));
  }
  for (i = 0; i < 8; i++) {
    dev_write(1, 0x0f * i / (8 - 1));
  }
  for (i = 0; i < 4; i++) {
    dev_write(1, 0x0f * i / (4 - 1));
  }
  dev_flush();

  // RAM Setting
  dev_write(0, CMD_PASET);
  dev_write(1, 0);
  dev_write(1, GLCD_HEIGHT - 1);
  dev_write(0, CMD_CASET);
  dev_write(1, 0);
  dev_write(1, GLCD_WIDTH - 1);
  dev_flush();

  // RAM Write
  dev_write(0, CMD_RAMWR);
  for (i = 0; i < GLCD_WIDTH * GLCD_HEIGHT; i++) {
    dev_write(1, 0x00);
  }
  dev_flush();

  // Waiting
  usleep(100000);
  dev_write(0, CMD_DISON);
  dev_flush();

  return 0;
}


void glcd_close(void) {
  dev_write(0, CMD_DISOFF);
  dev_flush();

  dev_write(0, CMD_PWRCTR);
  dev_write(1, 0x00);
  dev_flush();

  dev_write(0, CMD_OSCOFF);
  dev_flush();

  dev_write(0, CMD_SLPIN);
  dev_flush();

  ftdi_disable_bitbang(&Ftdic);
  ftdi_usb_close(&Ftdic);
  ftdi_deinit(&Ftdic);

  return;
}


void glcd_draw(int x, int y, int w, int h, const unsigned char *bmp) {
  int i;

  if (x < 0 || x >= GLCD_WIDTH ||
      y < 0 || y >= GLCD_HEIGHT ||
      w < 0 || x + w > GLCD_WIDTH ||
      h < 0 || y + h > GLCD_HEIGHT) return;

  dev_write(0, CMD_PASET);
  dev_write(1, y);
  dev_write(1, y + h - 1);
  dev_write(0, CMD_CASET);
  dev_write(1, x);
  dev_write(1, x + w - 1);
  dev_flush();

  dev_write(0, CMD_RAMWR);
  for (i = 0; i < w * h; i++) {
    dev_write(1, bmp[i]);
  }
  dev_flush();

  return;
}


int glcd_getc() {
  unsigned char pins;

  if (ftdi_read_pins(&Ftdic, &pins) < 0) {
    return -1;
  }
  if ((pins & 0x10) == 0) {
    return 0;
  } else if ((pins & 0x20) == 0) {
    return 2;
  } else if ((pins & 0x40) == 0) {
    return 3;
  } else if ((pins & 0x80) == 0) {
    return 1;
  } else {
    return -1;
  }
}
