/*
  Graphic LCD NOKIA3300-LCD Daemon

  T. Nakagawa
  2014/01/09
*/


#define _XOPEN_SOURCE 600
#define _BSD_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <iconv.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>
#include "glcd.h"
#include "misaki.xbm"
#include "msk_4x8.xbm"


#define IDENT "lcdtermd"
#define KEY_SENSE_INTVL 30	/* Frequency of key sense per second */
#define KEY_REPT_WAIT 500	/* Time to start key repeat [msec] */
#define KEY_REPT_FREQ 15	/* Frequency of key repeat per second */
#define SCR_WIDTH 32
#define SCR_HEIGHT 16
#define BUF_SIZE 1024


int Mfd;
iconv_t Iconv;
struct {
  int x;
  int y;
  int fgcol;
  int bgcol;
  unsigned char vram[GLCD_WIDTH * GLCD_HEIGHT];
  char dirty[GLCD_HEIGHT];
  int scan_line;
} Scr;
unsigned char Pal[8] = {
  0x00,	/* Black */
  0xe0,	/* Red */
  0x1c,	/* Green */
  0xfc,	/* Yellow */
  0x03,	/* Blue */
  0xe3,	/* Magenta */
  0x1f,	/* Cyan */
  0xff,	/* White */
};
char Key[4] = {'h', 'j', 'k', 'l'};


void scr_put_ascii(unsigned char c) {
  int i, j;
  int row, col;
  int ptr;
  int pos;

  if (c < 0x20) return;
  row = c / 16;
  col = c % 16;
  ptr = row * 8 * msk_4x8_width + col * 4;
  pos = (2 + Scr.y * 8) * GLCD_WIDTH + Scr.x * 4;
  for (i = 0; i < 8; i++) {
    unsigned char pat;
    if (ptr < 0 || ptr >= msk_4x8_width * msk_4x8_height) return;
    if (pos < 0 || pos >= GLCD_WIDTH * GLCD_HEIGHT) return;
    pat = msk_4x8_bits[ptr / 8];
    if (col % 2 == 1) {
      pat >>= 4;
    }
    for (j = 0; j < 4; j++) {
      Scr.vram[pos + j] = Pal[((pat & (0x01 << j)) != 0) ? Scr.fgcol : Scr.bgcol];
    }
    Scr.dirty[2 + Scr.y * 8 + i] = 1;
    ptr += msk_4x8_width;
    pos += GLCD_WIDTH;
  }
}


void scr_put_eucjp(unsigned short c) {
  int i, j;
  int row, col;
  int ptr;
  int pos;

  row = c / 256 - 0xa1;
  col = c % 256 - 0xa1;
  ptr = row * 8 * misaki_width + col * 8;
  pos = (2 + Scr.y * 8) * GLCD_WIDTH + Scr.x * 4;
  for (i = 0; i < 8; i++) {
    unsigned char pat;
    if (ptr < 0 || ptr >= misaki_width * misaki_height) return;
    if (pos < 0 || pos >= GLCD_WIDTH * GLCD_HEIGHT) return;
    pat = misaki_bits[ptr / 8];
    for (j = 0; j < 8; j++) {
      Scr.vram[pos + j] = Pal[((pat & (0x01 << j)) != 0) ? Scr.fgcol : Scr.bgcol];
    }
    Scr.dirty[2 + Scr.y * 8 + i] = 1;
    ptr += misaki_width;
    pos += GLCD_WIDTH;
  }
}


void scr_clear(int x, int y) {
  int i ,j;
  int pos;

  pos = (2 + y * 8) * GLCD_WIDTH + x * 4;
  for (i = 0; i < 8; i++) {
    if (pos < 0 || pos >= GLCD_WIDTH * GLCD_HEIGHT) return;
    for (j = 0; j < 4; j++) {
      Scr.vram[pos + j] = Pal[0];
    }
    Scr.dirty[2 + y * 8 + i] = 1;
    pos += GLCD_WIDTH;
  }
}


void scr_scroll(void) {
  int i;

  memmove(Scr.vram + 2 * GLCD_WIDTH, Scr.vram + (2 + 8) * GLCD_WIDTH, sizeof(unsigned char) * GLCD_WIDTH * 8 * (SCR_HEIGHT - 1));
  for (i = 0; i < 8 * (SCR_HEIGHT - 1); i++) {
    Scr.dirty[2 + i] = 1;
  }
  for (i = 0; i < SCR_WIDTH; i++) {
    scr_clear(i, SCR_HEIGHT - 1);
  }
}


void scr_update(int line) {
  glcd_draw(0, line, GLCD_WIDTH, 1, Scr.vram + line * GLCD_WIDTH);
  Scr.dirty[line] = 0;
}


void process(char c) {
  int i, j;
  static char cc = '\0';

  if (cc != '\0') {
    if (cc == '\x01') {
      /* Cursor X position. */
      c -= 32;
      if (c >= 0 && c < SCR_WIDTH) {
        Scr.x = c;
      }
    } else if (cc == '\x02') {
      /* Cursor Y position; */
      c -= 32;
      if (c >= 0 && c < SCR_HEIGHT) {
        Scr.y = c;
      }
    } else if (cc == '\x03') {
      /* Foreground color. */
      c -= 32;
      if (c >= 0 && c < 8) {
        Scr.fgcol = c;
      }
    } else if (cc == '\x04') {
      /* Background color. */
      c -= 32;
      if (c >= 0 && c < 8) {
        Scr.bgcol = c;
      }
    } else if (cc == '\x11') {
      Key[0] = c;
    } else if (cc == '\x12') {
      Key[1] = c;
    } else if (cc == '\x13') {
      Key[2] = c;
    } else if (cc == '\x14') {
      Key[3] = c;
    }
    cc = '\0';
  } else {
    if (c == '\x01' ||
        c == '\x02' ||
        c == '\x03' ||
        c == '\x04' ||
        c == '\x11' ||
        c == '\x12' ||
        c == '\x13' ||
        c == '\x14') {
      cc = c;	/* The next character is necessary to process. */
    } else if (c == '\x08') {	/* ^H */
      /* Backspace */
      if (Scr.x >= 1) {
        Scr.x--;
      }
    } else if (c == '\x0a') {	/* ^J */
      /* Line feed */
      Scr.y++;
      if (Scr.y >= SCR_HEIGHT) {
        Scr.y--;
        scr_scroll();
      }
    } else if (c == '\x0b') {
      /* Erase the rest of the line. */
      for (i = Scr.x; i < SCR_WIDTH; i++) {
        scr_clear(i, Scr.y);
      }
    } else if (c == '\x0d') {
      /* Carriage return. */
      Scr.x = 0;
    } else if (c == '\x0e') {
      /* Erase the rest of the screen. */
      for (i = Scr.x; i < SCR_WIDTH; i++) {
        scr_clear(i, Scr.y);
      }
      for (i = Scr.y + 1; i < SCR_HEIGHT; i++) {
        for (j = 0; j < SCR_WIDTH; j++) {
          scr_clear(j, i);
        }
      }
    } else if ((c & 0x80) != 0) {
      /* UTF-8 character */
      static char ibuf[3];
      static size_t isize = 0;
      char obuf[2];
      size_t osize = 2;
      char *iptr = ibuf, *optr = obuf;
      if (isize >= 3) {
        ibuf[0] = ibuf[1];
        ibuf[1] = ibuf[2];
        isize--;
      }
      ibuf[isize++] = c;
      if (iconv(Iconv, &iptr, &isize, &optr, &osize) < 0) {
        iconv(Iconv, NULL, NULL, NULL, NULL);
        isize = 0;
      }
      if (osize == 0) {
        /* EUC character */
        if (Scr.x >= SCR_WIDTH - 1) {
          Scr.x = 0;
          Scr.y++;
          if (Scr.y >= SCR_HEIGHT) {
            Scr.y--;
            scr_scroll();
          }
        }
        scr_put_eucjp(((unsigned char)obuf[0] << 8) | (unsigned char)obuf[1]);
        Scr.x += 2;
        isize = 0;
      }
    } else {
      /* ASCII character */
      if (Scr.x >= SCR_WIDTH) {
        Scr.x = 0;
        Scr.y++;
        if (Scr.y >= SCR_HEIGHT) {
          Scr.y--;
          scr_scroll();
        }
      }
      scr_put_ascii(c);
      Scr.x += 1;
    }
  }

  return;
}


void sighandler(int sig) {
  if (sig == SIGTERM) {
    close(Mfd);
    iconv_close(Iconv);
    glcd_close();
    syslog(LOG_INFO, "Terminated");
    exit(0);
  } else if (sig == SIGALRM) {
    static int prvkey = -1;
    static int count = 0;
    int key;

    key = glcd_getc();
    if (key >= 0) {
      const char buf = Key[key];
      if (key == prvkey) {
        if (--count == 0) {
          write(Mfd, &buf, sizeof(char));
          count = KEY_SENSE_INTVL / KEY_REPT_FREQ;
        }
      } else {
        write(Mfd, &buf, sizeof(char));
        count = KEY_SENSE_INTVL * KEY_REPT_WAIT / 1000;
      }
    }
    prvkey = key;
  }

  return;
}


int main(int argc, char **argv) {
  int i;
  struct sigaction sa;
  struct itimerval it;
  int flags;
  char buf[BUF_SIZE];
  int size;
  char *ptr;
  int pid;
  int sfd;
  struct winsize winsz;

  Scr.x = 0;
  Scr.y = 0;
  Scr.fgcol = 7;
  Scr.bgcol = 0;
  Scr.scan_line = 0;

  openlog(IDENT, (LOG_CONS | LOG_PERROR | LOG_PID), LOG_DAEMON);

  if ((Mfd = posix_openpt(O_RDWR)) < 0 ||
      grantpt(Mfd) < 0 ||
      unlockpt(Mfd) < 0) {
    syslog(LOG_ERR, "cannot open pseudo tty (master)");
    return 1;
  }
  if ((sfd = open(ptsname(Mfd), O_RDWR)) < 0) {
    syslog(LOG_ERR, "cannot open pseudo tty (slave)");
    return 1;
  }
  winsz.ws_row = SCR_HEIGHT;
  winsz.ws_col = SCR_WIDTH;
  ioctl(sfd, TIOCSWINSZ, &winsz);
  syslog(LOG_INFO, "Initialized the pseudo tty: %s.", ptsname(Mfd));

  if ((pid = fork()) < 0) {
    syslog(LOG_ERR, "fork() failed");
    return 1;
  }
  if (pid == 0) {
    close(Mfd);
    close(0);
    close(1);
    close(2);
    dup(sfd);
    dup(sfd);
    dup(sfd);
    close(sfd);
    setsid();
    ioctl(0, TIOCSCTTY, 1);
    if (argc >= 2) {
      execvp(argv[1], argv + 1);
      syslog(LOG_ERR, "failed to launch a child process");
      return 1;
    }
    return 0;
  }

  if ((Iconv = iconv_open("EUC-JP-MS", "UTF-8")) < 0) {
    syslog(LOG_ERR, "iconv_open() failed");
    return 1;
  }

  if (glcd_open()) {
    syslog(LOG_ERR, "glcd_open() failed.");
    return 1;
  }

  sa.sa_handler = sighandler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGTERM, &sa, NULL) < 0 || sigaction(SIGALRM, &sa, NULL) < 0) {
    syslog(LOG_ERR, "sigaction() failed");
    return 1;
  }

  it.it_interval.tv_sec = it.it_value.tv_sec = 0;
  it.it_interval.tv_usec = it.it_value.tv_usec = 1000000 / KEY_SENSE_INTVL;
  if (setitimer(ITIMER_REAL, &it, NULL) < 0) {
    syslog(LOG_ERR, "setitimer() failed");
    return 1;
  }

  syslog(LOG_INFO, "Started.");

  flags = fcntl(Mfd, F_GETFL, 0);
  size = 0;
  for (; ; ) {
    if (size > 0) {
      process(*ptr++);
      size--;
      continue;
    }

    fcntl(Mfd, F_SETFL, flags | O_NONBLOCK);
    size = read(Mfd, buf, BUF_SIZE);
    ptr = buf;
    if (size == 0) {
      break;
    } else if (size > 0) {
      continue;
    } else if (errno != EAGAIN) {
      syslog(LOG_ERR, "nonblocking read() failed: %s", strerror(errno));
      return 1;
    }

    for (i = 0; i < GLCD_HEIGHT; i++) {
      Scr.scan_line++;
      if (Scr.scan_line >= GLCD_HEIGHT) {
        Scr.scan_line = 0;
      }
      if (Scr.dirty[Scr.scan_line]) break;
    }
    if (i < GLCD_HEIGHT) {
      scr_update(Scr.scan_line);
      continue;
    }

    fcntl(Mfd, F_SETFL, flags & ~O_NONBLOCK);
    size = read(Mfd, buf, BUF_SIZE);
    ptr = buf;
    if (size == 0) {
      break;
    } else if (size < 0) {
      syslog(LOG_ERR, "blocking read() failed: %s", strerror(errno));
      return 1;
    }
  }

  return 0;
}
